﻿// PianoKeyboard by Charles Petzold, January 2010
using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;

namespace Piano
{
    public partial class PianoKeyboard : UserControl
    {
        const double WHITE_KEY_WIDTH = 90;
        const double WHITE_KEY_HEIGHT = 400;
        const double BLACK_KEY_WIDTH = 0.6 * WHITE_KEY_WIDTH;
        const double BLACK_KEY_HEIGHT = 0.65 * WHITE_KEY_HEIGHT;

        static Brush whiteBrush = new SolidColorBrush(Colors.White);
        static Brush blackBrush = new SolidColorBrush(Colors.Black);

        static bool[] isBlack = { false, true, false, true, 
                                  false, false, true, false, 
                                  true, false, true, false };

        static int[] blackIndex = { 0, 0, 0, 1, 0, 0, 2, 0, 3, 0, 4, 0 };
        static double[] blackOffsets = { 0.4, 0.2, 0.4, 0.3, 0.2 };

        List<PianoKey> sustainedKeys = new List<PianoKey>();

        public static readonly DependencyProperty IsSustainPedalDownProperty =
            DependencyProperty.Register("IsSustainPedalDown",
                typeof(bool),
                typeof(PianoKeyboard),
                new PropertyMetadata(false, OnIsSustainPedalDownChanged));

        public static readonly DependencyProperty VolumeProperty =
            DependencyProperty.Register("Volume",
                typeof(double),
                typeof(PianoKeyboard),
                new PropertyMetadata(0.0));

        public PianoKeyboard()
        {
            InitializeComponent();
        }

        public PianoVoices PianoVoices { set; get; }

        public bool IsSustainPedalDown
        {
            set { SetValue(IsSustainPedalDownProperty, value); }
            get { return (bool)GetValue(IsSustainPedalDownProperty); }
        }

        public double Volume
        {
            set { SetValue(VolumeProperty, value); }
            get { return (double)GetValue(VolumeProperty); }
        }

        public void Build(int firstKey, int lastKey)
        {
            double offset = 0;

            for (int keyCode = firstKey; keyCode <= lastKey; keyCode++)
            {
                bool isBlackKey = isBlack[keyCode % 12];
                
                PianoKey key = new PianoKey()
                {
                    Width = isBlackKey ? BLACK_KEY_WIDTH : WHITE_KEY_WIDTH,
                    Height = isBlackKey ? BLACK_KEY_HEIGHT : WHITE_KEY_HEIGHT,
                    Background = isBlackKey ? blackBrush : whiteBrush,
                    KeyCode = keyCode,
                    IsMiddleC = keyCode == 60,
                    VerticalAlignment = VerticalAlignment.Top
                };

                key.Pressed += OnKeyPressed;
                key.Released += OnKeyReleased;

                if (isBlackKey)
                {
                    Canvas.SetLeft(key, offset - WHITE_KEY_WIDTH * blackOffsets[blackIndex[keyCode % 12]]);
                    canvas.Children.Add(key);
                }
                else
                {
                    stackPanel.Children.Add(key);
                }
                
                offset += isBlackKey ? 0 : WHITE_KEY_WIDTH;
            }

        }

        void OnKeyPressed(object sender, EventArgs args)
        {
            PianoKey pianoKey = sender as PianoKey;
            PianoVoices.TriggerVoice(pianoKey.KeyCode, pianoKey, Volume);
        }

        void OnKeyReleased(object sender, EventArgs args)
        {
            PianoKey pianoKey = sender as PianoKey;

            if (IsSustainPedalDown)
            {
                sustainedKeys.Add(pianoKey);
            }
            // Only damp if sustain pedal is up
            else
            {
                PianoVoices.DampVoice(pianoKey);
            }
        }

        static void OnIsSustainPedalDownChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            (obj as PianoKeyboard).OnIsSustainPedalDownChanged(args);
        }

        void OnIsSustainPedalDownChanged(DependencyPropertyChangedEventArgs args)
        {
            // If Sustain pedal being lifted, damp all sustained notes
            if (!(bool)args.NewValue)
            {
                foreach (PianoKey pianoKey in sustainedKeys)
                {
                    PianoVoices.DampVoice(pianoKey);
                }
                sustainedKeys.Clear();
            }
        }
    }
}
