﻿// TouchDial by Charles Petzold, January 2010
using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Input;
using System.ComponentModel;
using System.Windows.Media;

namespace Petzold.MultiTouch
{
    public class TouchDial : RangeBase
    {
        TouchManager touchManager;
        Point point0;
        bool isCaptured;
        double nonSnappedValue;

        public static readonly DependencyProperty RotateTransformProperty =
            DependencyProperty.Register("RotateTransform",
                typeof(RotateTransform),
                typeof(TouchDial),
                new PropertyMetadata(null));

        public static readonly DependencyProperty InputCenterXProperty =
            DependencyProperty.Register("InputCenterX",
                typeof(double),
                typeof(TouchDial),
                new PropertyMetadata(Double.NaN));

        public static readonly DependencyProperty InputCenterYProperty =
            DependencyProperty.Register("InputCenterY",
                typeof(double),
                typeof(TouchDial),
                new PropertyMetadata(Double.NaN));

        public static readonly DependencyProperty RenderCenterXProperty =
            DependencyProperty.Register("RenderCenterX",
                typeof(double),
                typeof(TouchDial),
                new PropertyMetadata(Double.NaN, OnRenderCenterChanged));

        public static readonly DependencyProperty RenderCenterYProperty =
            DependencyProperty.Register("RenderCenterY",
                typeof(double),
                typeof(TouchDial),
                new PropertyMetadata(Double.NaN, OnRenderCenterChanged));

        public static readonly DependencyProperty SnapIncrementProperty =
            DependencyProperty.Register("SnapIncrement",
                typeof(double),
                typeof(TouchDial),
                new PropertyMetadata(0.0));

        public TouchDial()
        {
            RotateTransform = new RotateTransform();
            SizeChanged += OnSizeChanged;

            touchManager = new TouchManager(this);
            touchManager.TouchDown += OnTouchDown;
            touchManager.TouchMove += OnTouchMove;
            touchManager.TouchUp += OnTouchUp;
            touchManager.LostTouchCapture += OnLostTouchCapture;
        }

        // Properties backed by dependency properties
        public RotateTransform RotateTransform
        {
            protected set { SetValue(RotateTransformProperty, value); }
            get { return (RotateTransform)GetValue(RotateTransformProperty); }
        }

        public double InputCenterX
        {
            set { SetValue(InputCenterXProperty, value); }
            get { return (double)GetValue(InputCenterXProperty); }
        }

        public double InputCenterY
        {
            set { SetValue(InputCenterYProperty, value); }
            get { return (double)GetValue(InputCenterYProperty); }
        }

        public double RenderCenterX
        {
            set { SetValue(RenderCenterXProperty, value); }
            get { return (double)GetValue(RenderCenterXProperty); }
        }

        public double RenderCenterY
        {
            set { SetValue(RenderCenterYProperty, value); }
            get { return (double)GetValue(RenderCenterYProperty); }
        }

        public double SnapIncrement
        {
            set { SetValue(SnapIncrementProperty, value); }
            get { return (double)GetValue(SnapIncrementProperty); }
        }

        // Properties for internal convenience
        protected double ActualInputCenterX
        {
            get { return Double.IsNaN(InputCenterX) ? ActualWidth / 2 : InputCenterX; }
        }

        protected double ActualInputCenterY
        {
            get { return Double.IsNaN(InputCenterY) ? ActualHeight / 2 : InputCenterY; }
        }

        protected double ActualRenderCenterX
        {
            get { return Double.IsNaN(RenderCenterX) ? ActualWidth / 2 : RenderCenterX; }
        }

        protected double ActualRenderCenterY
        {
            get { return Double.IsNaN(RenderCenterY) ? ActualHeight / 2 : RenderCenterY; }
        }

        // Changed handlers
        void OnSizeChanged(object sender, SizeChangedEventArgs args)
        {
            CalculateNewRotateTransform();
        }

        static void OnRenderCenterChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            (obj as TouchDial).CalculateNewRotateTransform();
        }

        protected override void OnValueChanged(double oldValue, double newValue)
        {
            CalculateNewRotateTransform();
            base.OnValueChanged(oldValue, newValue);
        }

        // Mouse processing
        protected override void OnMouseLeftButtonDown(MouseButtonEventArgs args)
        {
            nonSnappedValue = Value;
            point0 = args.GetPosition(this);
            CaptureMouse();
            isCaptured = true;
            args.Handled = true;
            base.OnMouseLeftButtonDown(args);
        }

        protected override void OnMouseMove(MouseEventArgs args)
        {
            if (isCaptured)
            {
                Point point1 = args.GetPosition(this);
                CrankDial(point1);
                point0 = point1;
            }
            base.OnMouseMove(args);
        }

        protected override void OnMouseLeftButtonUp(MouseButtonEventArgs args)
        {
            if (isCaptured)
            {
                Point point1 = args.GetPosition(this);
                CrankDial(point1);
                point0 = point1;
                args.Handled = true;
            }
            base.OnMouseLeftButtonUp(args);
        }

        protected override void OnLostMouseCapture(MouseEventArgs args)
        {
            isCaptured = false;
            base.OnLostMouseCapture(args);
        }

        // Touch processing
        void OnTouchDown(object sender, TouchEventArgs args)
        {
            // Don't allow multiple touch points for this one control
            if (isCaptured)
                return;

            nonSnappedValue = Value;
            point0 = args.Position;
            touchManager.Capture(args.Id);
            isCaptured = true;
        }

        void OnTouchMove(object sender, TouchEventArgs args)
        {
            if (isCaptured)
            {
                Point point1 = args.Position;
                CrankDial(point1);
                point0 = point1;
            }
        }

        void OnTouchUp(object sender, TouchEventArgs args)
        {
            if (isCaptured)
            {
                Point point1 = args.Position;
                CrankDial(point1);
                point0 = point1;
            }
        }

        void OnLostTouchCapture(object sender, TouchEventArgs arge)
        {
            isCaptured = false;
        }

        // Basic calcuation to change Value based on mouse or finger movement
        void CrankDial(Point point1)
        {
            double radians0 = Math.Atan2(point0.Y - ActualInputCenterY, 
                                         point0.X - ActualInputCenterX);

            double radians1 = Math.Atan2(point1.Y - ActualInputCenterY, 
                                         point1.X - ActualInputCenterX);

            double angleDiff = 180 * (radians1 - radians0) / Math.PI;

            while (Math.Abs(angleDiff) > 180)
                angleDiff -= Math.Sign(angleDiff) * 360;

            nonSnappedValue += angleDiff;
            Value = (SnapIncrement == 0) ? nonSnappedValue : SnapIncrement * Math.Round(nonSnappedValue / SnapIncrement);
        }

        // Calculation of RotateTransform property 
        void CalculateNewRotateTransform()
        {
            RotateTransform rotate = RotateTransform;
            rotate.Angle = Value;
            rotate.CenterX = ActualRenderCenterX;
            rotate.CenterY = ActualRenderCenterY;
            RotateTransform = rotate;
        }
    }
}
