﻿// TouchManager by Charles Petzold, January 2010
using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Input;
using System.Windows.Media;

namespace Petzold.MultiTouch
{
    public class TouchManager
    {
        static Dictionary<UIElement, TouchManager> elementDictionary = new Dictionary<UIElement, TouchManager>();
        static Dictionary<int, TouchManager> captureDictionary = new Dictionary<int, TouchManager>();
        static Dictionary<int, TouchManager> lastMoveDictionary = new Dictionary<int, TouchManager>();

        static TouchManager()
        {
            Touch.FrameReported += OnTouchFrameReported;
        }

        static void OnTouchFrameReported(object sender, TouchFrameEventArgs args)
        {
            // Get the primary touch point to possibly suspend mouse promotion
            TouchPoint primaryTouchPoint = args.GetPrimaryTouchPoint(null);

            if (primaryTouchPoint != null && primaryTouchPoint.Action == TouchAction.Down)
            {
                UIElement child = primaryTouchPoint.TouchDevice.DirectlyOver;

                if (child != null)
                {
                    foreach (UIElement ancestor in elementDictionary.Keys)
                    {
                        if (IsAncestorOf(ancestor, child))
                        {
                            args.SuspendMousePromotionUntilTouchUp();
                            break;
                        }
                    }
                }
            }

            // Get all touch points and enumerate through them
            TouchPointCollection touchPoints = args.GetTouchPoints(null);

            foreach (TouchPoint touchPoint in touchPoints)
            {
                // Get objects associated with touch point
                TouchDevice touchDevice = touchPoint.TouchDevice;
                int id = touchDevice.Id;
                UIElement directlyOverElement = touchDevice.DirectlyOver;

                if (touchPoint.Action == TouchAction.Down && directlyOverElement == null)
                    continue;

                // --------------------------
                // Touch point ID is captured
                // --------------------------
                if (captureDictionary.ContainsKey(id))
                {
                    TouchManager touchManager = captureDictionary[id];
                    
                    // Create TouchEventArgs for event
                    TouchEventArgs touchArgs = new TouchEventArgs()
                    {
                        Id = id,
                        Source = touchManager.sourceElement,
                        Position = TransformPoint(touchManager.sourceElement, touchPoint.Position),
                        DirectlyOver = directlyOverElement
                    };

                    switch (touchPoint.Action)
                    {
                        case TouchAction.Down:
                            throw new Exception("Not expecting TouchAction.Down event when touch point is captured");

                        case TouchAction.Move:
                            touchManager.OnTouchMove(touchArgs);
                            break;

                        case TouchAction.Up:
                            touchManager.OnTouchUp(touchArgs);
                            touchManager.OnTouchLeave(touchArgs);
                            touchManager.OnLostTouchCapture(touchArgs);
                            captureDictionary.Remove(id);

                            if (lastMoveDictionary.ContainsKey(id))
                                lastMoveDictionary.Remove(id);

                            break;
                    }
                }
                // ------------------------------
                // Touch point ID is not captured
                // ------------------------------
                else
                {
                    UIElement testElement = directlyOverElement;
                    bool newCapture = false;
                    bool generatedEvent = false;

                    while (!newCapture && testElement != null)
                    {
                        if (elementDictionary.ContainsKey(testElement))
                        {
                            TouchManager touchManager = elementDictionary[testElement];

                            // Create TouchEventArgs for eventual event
                            TouchEventArgs touchArgs = new TouchEventArgs()
                            {
                                Id = id,
                                Source = touchManager.sourceElement,
                                Position = TransformPoint(touchManager.sourceElement, touchPoint.Position),
                                DirectlyOver = directlyOverElement
                            };

                            switch (touchPoint.Action)
                            {
                                case TouchAction.Down:
                                    touchManager.OnTouchEnter(touchArgs);

                                    touchManager.hasMadeNewCapture = false;
                                    touchManager.isInTouchDown = true;
                                    touchManager.OnTouchDown(touchArgs);
                                    touchManager.isInTouchDown = false;
                                    lastMoveDictionary.Add(id, touchManager);

                                    // If there's a new capture, then want to inhibit further searching up the tree
                                    if (touchManager.hasMadeNewCapture)
                                        newCapture = true;

                                    generatedEvent = true;
                                    break;

                                case TouchAction.Move:
                                    if (lastMoveDictionary.ContainsKey(id))
                                    {
                                        TouchManager lastMoveTouchManager = lastMoveDictionary[id];

                                        if (lastMoveTouchManager != touchManager)
                                        {
                                            TouchEventArgs lastMoveTouchArgs = new TouchEventArgs()
                                            {
                                                Id = id,
                                                Source = lastMoveTouchManager.sourceElement,
                                                Position = TransformPoint(lastMoveTouchManager.sourceElement, touchPoint.Position),
                                                DirectlyOver = directlyOverElement
                                            };

                                            lastMoveTouchManager.OnTouchLeave(lastMoveTouchArgs);
                                            touchManager.OnTouchEnter(touchArgs);
                                            lastMoveDictionary[touchDevice.Id] = touchManager;
                                        }
                                    }
                                    else
                                    {
                                        touchManager.OnTouchEnter(touchArgs);
                                        lastMoveDictionary.Add(id, touchManager);
                                    }
                                    
                                    touchManager.OnTouchMove(touchArgs);
                                    generatedEvent = true;
                                    break;

                                case TouchAction.Up:
                                    touchManager.OnTouchUp(touchArgs);
                                    touchManager.OnTouchLeave(touchArgs);
                                    lastMoveDictionary.Remove(id);
                                    generatedEvent = true;
                                    break;
                            }
                        }
                        // Go next element up the tree for the next iteration
                        testElement = VisualTreeHelper.GetParent(testElement) as UIElement;
                    }

                    // It's possible that a Move has gone beyond the area of an 
                    //      element registered with TouchManager. This code is
                    //      necessary for TouchLeave to work  
                    if (!generatedEvent &&
                            (touchPoint.Action == TouchAction.Move || touchPoint.Action == TouchAction.Up) &&
                            lastMoveDictionary.ContainsKey(id))
                    {
                        TouchManager lastMoveTouchManager = lastMoveDictionary[touchDevice.Id];
                        TouchEventArgs lastMoveTouchArgs = new TouchEventArgs()
                        {
                            Id = id,
                            Source = lastMoveTouchManager.sourceElement,
                            Position = TransformPoint(lastMoveTouchManager.sourceElement, touchPoint.Position),
                            DirectlyOver = directlyOverElement
                        };

                        lastMoveTouchManager.OnTouchLeave(lastMoveTouchArgs);
                        lastMoveDictionary.Remove(touchDevice.Id);
                    }
                }
            }
        }

        static bool IsAncestorOf(DependencyObject ancestor, DependencyObject child)
        {
            if (ancestor == child)
                return true;

            DependencyObject parent = VisualTreeHelper.GetParent(child);

            if (parent == null)
                return false;

            return IsAncestorOf(ancestor, parent);
        }

        static Point TransformPoint(UIElement sourceElement, Point point)
        {
            return sourceElement.TransformToVisual(null).Inverse.Transform(point);
        }

        //-------------------------------------------------------
        // Division between static and instance portion of class
        //-------------------------------------------------------

        UIElement sourceElement;
        bool isInTouchDown;
        bool hasMadeNewCapture;

        public event EventHandler<TouchEventArgs> TouchEnter;
        public event EventHandler<TouchEventArgs> TouchDown;
        public event EventHandler<TouchEventArgs> TouchMove;
        public event EventHandler<TouchEventArgs> TouchUp;
        public event EventHandler<TouchEventArgs> TouchLeave;
        public event EventHandler<TouchEventArgs> LostTouchCapture;

        public TouchManager(UIElement sourceElement)
        {
            this.sourceElement = sourceElement;
            elementDictionary.Add(sourceElement, this);
        }

        public bool Capture(int id)
        {
            if (!isInTouchDown)
                return false;               // should also check ID is the same???

            TouchManager capturedTouchManager;

            if (captureDictionary.TryGetValue(id, out capturedTouchManager))
            {
                return capturedTouchManager.sourceElement == sourceElement;
            }

            captureDictionary.Add(id, this);
            hasMadeNewCapture = true;
            return true;
        }

        public void ReleaseCapture(int id)
        {
            if (captureDictionary.ContainsKey(id))
            {
                TouchManager capturedTouchManager = captureDictionary[id];

                TouchEventArgs args = new TouchEventArgs()
                {
                     Source = capturedTouchManager.sourceElement,
                     Id = id
                };
                OnLostTouchCapture(args);

                captureDictionary.Remove(id);
            }
        }

        void OnTouchEnter(TouchEventArgs args)
        {
            if (TouchEnter != null)
                TouchEnter(this, args);
        }

        void OnTouchDown(TouchEventArgs args)
        {
            if (TouchDown != null)
                TouchDown(this, args);
        }

        void OnTouchMove(TouchEventArgs args)
        {
            if (TouchMove != null)
                TouchMove(this, args);
        }

        void OnTouchUp(TouchEventArgs args)
        {
            if (TouchUp != null)
                TouchUp(this, args);
        }

        void OnTouchLeave(TouchEventArgs args)
        {
            if (TouchLeave != null)
                TouchLeave(this, args);
        }

        void OnLostTouchCapture(TouchEventArgs args)
        {
            if (LostTouchCapture != null)
                LostTouchCapture(this, args);
        }
    }
}
