﻿// FmPiano by Charles Petzold, January 2010
// Based on Bill Schottstaedt, "The Simulation of Natural Instrument Tones Using Frequency Modulation with a Complex Modulating Wave,"
//      Computer Music Journal, Vol 1, No. 4 (1977), 46-50.
//      Republished in Curtis Roads & John Strawn, ed., Foundations of Computer Music (MIT Press, 1985), 54-64.

using System;

namespace Petzold.SoundSynthesis
{
    public class FmPiano : ISampleProvider
    {
        SineWave sineWaveCarrier;
        SineWave sineWaveModulator1;
        SineWave sineWaveModulator2;
        Adder modulatorsAdder = new Adder();
        Adder frequencyAdder = new Adder();
        SegmentEnvelope amplitudeEnvelope;

        public event EventHandler EnvelopeCompleted;

        public FmPiano(int sampleRate)
        {
            // Create sine wave oscillators
            sineWaveCarrier = new SineWave(sampleRate);
            sineWaveModulator1 = new SineWave(sampleRate);
            sineWaveModulator2 = new SineWave(sampleRate);

            // Set inputs of modulatorsAdder
            modulatorsAdder.Input1 = sineWaveModulator1;
            modulatorsAdder.Input2 = sineWaveModulator2;
            
            // Set inputs of frequencyAdder
            frequencyAdder.Input1 = modulatorsAdder;
            // frequencyAdder.Input2 is set in Attack() method

            // Create and initialize envelope
            amplitudeEnvelope = new SegmentEnvelope(sampleRate);
            amplitudeEnvelope.AddSegment(0.00, 0.00);
            amplitudeEnvelope.AddSegment(0.00005, 1.00);
            amplitudeEnvelope.AddSegment(0.05, 0.60);
            amplitudeEnvelope.AddSegment(0.10, 0.30);
            amplitudeEnvelope.AddSegment(0.25, 0.15);
            amplitudeEnvelope.AddSegment(0.50, 0.07);
            amplitudeEnvelope.AddSegment(1.00, 0.00);
            amplitudeEnvelope.EnvelopeCompleted += OnEnvelopeCompleted;

            // Set inputs of sineWaveCarrier
            sineWaveCarrier.Frequency = frequencyAdder;
            sineWaveCarrier.Amplitude = amplitudeEnvelope;
        }

        public double Frequency { set; get; }
        public double Amplitude { get; set; }

        public void Attack(double volume)
        {
            // Calculate values
            double frequencyLog = Math.Log(Frequency);
            double logFlip = Math.Max(0, 8 - frequencyLog);
            double indexOfModulation1 = 17 * logFlip / Math.Pow(frequencyLog, 2);
            double indexOfModulation2 = 20 * logFlip / Frequency;
            double stretch = Frequency / 200;
            double modulatingFrequency1 = 1 * Frequency + stretch;
            double modulatingFrequency2 = 4 * Frequency + stretch;

            sineWaveModulator1.Frequency = new Constant(modulatingFrequency1);
            sineWaveModulator1.Amplitude = new Constant(modulatingFrequency1 * indexOfModulation1);

            sineWaveModulator2.Frequency = new Constant(modulatingFrequency2);
            sineWaveModulator2.Amplitude = new Constant(modulatingFrequency2 * indexOfModulation2);

            frequencyAdder.Input2 = new Constant(Frequency);

            // volume argument is in decibels, 0 to 96
            // amplitude multiplier is short, 0 to 32767
            short amplitudeMultiplier = (short) (Math.Pow(2, volume / 6 - 1) - 1);

            // peak amplitude for timing envelope is arbitrary scale to 2000
            double peakAmplitude = amplitudeMultiplier / 16.384;
            int timeMultiplier = (int)(10000 * Math.Sqrt(peakAmplitude) / Math.Sqrt(Frequency));

            amplitudeEnvelope.Attack(timeMultiplier, amplitudeMultiplier);
        }

        public void Damp()
        {
            amplitudeEnvelope.Damp(200);    // amplitude per msec, ie, about 32767/200 msec from maximum
        }

        public short GetNextSample()
        {
            return (short)(sineWaveCarrier.GetNextSample()); // IntSample() >> 16);
        }

        void OnEnvelopeCompleted(object sender, EventArgs args)
        {
            if (EnvelopeCompleted != null)
                EnvelopeCompleted(this, args);
        }

    }
}
