﻿// PcmMediaStreamSource.cs 
// adapted from http://community.irritatedvowel.com/blogs/pete_browns_blog/archive/2009/03/23/Creating-Sound-using-MediaStreamSource-in-Silverlight-3-Beta.aspx

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Windows.Media;

namespace Petzold.SoundSynthesis
{
    public class PcmMediaStreamSource : MediaStreamSource
    {
        const int BitsPerSample = 16;
        int sampleRate;
        int channelCount;
        int byteRate;
        short blockAlign;
        MediaStreamDescription mediaStreamDescription;
        long startPosition;
        long currentPosition;
        long currentTimeStamp;
        MemoryStream memoryStream;
        Dictionary<MediaSampleAttributeKeys, string> emptySampleDict = new Dictionary<MediaSampleAttributeKeys, string>();

        public PcmMediaStreamSource(int sampleRate, int channelCount, int audioBufferLength)
        {
            this.sampleRate = sampleRate;
            this.channelCount = channelCount;
            byteRate = sampleRate * channelCount * BitsPerSample / 8;
            blockAlign = (short)(channelCount * (BitsPerSample / 8));
            AudioBufferLength = audioBufferLength;
            memoryStream = new MemoryStream();
        }

        public ISampleProvider SampleProvider { get; set; }

        protected override void OpenMediaAsync()
        {
            startPosition = currentPosition = 0;

            Dictionary<MediaStreamAttributeKeys, string> streamAttributes = new Dictionary<MediaStreamAttributeKeys, string>();
            Dictionary<MediaSourceAttributesKeys, string> sourceAttributes = new Dictionary<MediaSourceAttributesKeys, string>();
            List<MediaStreamDescription> availableStreams = new List<MediaStreamDescription>();

            string format = "";
            format += ToLittleEndianString(string.Format("{0:X4}", 1));      // indicates PCM
            format += ToLittleEndianString(string.Format("{0:X4}", channelCount));
            format += ToLittleEndianString(string.Format("{0:X8}", sampleRate));
            format += ToLittleEndianString(string.Format("{0:X8}", byteRate));
            format += ToLittleEndianString(string.Format("{0:X4}", blockAlign));
            format += ToLittleEndianString(string.Format("{0:X4}", BitsPerSample));
            format += ToLittleEndianString(string.Format("{0:X4}", 0));

            streamAttributes[MediaStreamAttributeKeys.CodecPrivateData] = format;
            mediaStreamDescription = new MediaStreamDescription(MediaStreamType.Audio, streamAttributes);
            availableStreams.Add(mediaStreamDescription);
            sourceAttributes[MediaSourceAttributesKeys.Duration] = "0";
            sourceAttributes[MediaSourceAttributesKeys.CanSeek] = "false";
            ReportOpenMediaCompleted(sourceAttributes, availableStreams);
        }

        protected override void GetSampleAsync(MediaStreamType mediaStreamType)
        {
            int numSamples = channelCount * 256;
            int bufferByteCount = BitsPerSample / 8 * numSamples;

            // Hard-Coded for one channel
            for (int i = 0; i < numSamples; i++)
            {
                short sample = 0;
                
                if (SampleProvider != null)
                    sample = SampleProvider.GetNextSample();

                memoryStream.WriteByte((byte)(sample & 0xFF));
                memoryStream.WriteByte((byte)(sample >> 8));
            }

            // Send out the next sample
            MediaStreamSample mediaStreamSample = 
                new MediaStreamSample(mediaStreamDescription, memoryStream, currentPosition, 
                                      bufferByteCount, currentTimeStamp, emptySampleDict);

            // Move timestamp and position forward
            currentTimeStamp += bufferByteCount * 10000000L / byteRate;
            currentPosition += bufferByteCount;

            ReportGetSampleCompleted(mediaStreamSample);
        }

        protected override void SeekAsync(long seekToTime)
        {
            ReportSeekCompleted(seekToTime);
        }

        protected override void CloseMedia()
        {
            startPosition = currentPosition = 0;
            mediaStreamDescription = null;
        }

        protected override void GetDiagnosticAsync(MediaStreamSourceDiagnosticKind diagnosticKind)
        {
            throw new NotImplementedException();
        }

        protected override void SwitchMediaStreamAsync(MediaStreamDescription mediaStreamDescription)
        {
            throw new NotImplementedException();
        }

        string ToLittleEndianString(string bigEndianString)
        {
            StringBuilder builder = new StringBuilder();

            for (int i = 0; i < bigEndianString.Length; i += 2)
                builder.Insert(0, bigEndianString.Substring(i, 2));

            return builder.ToString();
        }
    }
}
