﻿// SegmentEnvelope by Charles Petzold, January 2010
using System;
using System.Collections.Generic;

namespace Petzold.SoundSynthesis
{
    public class SegmentEnvelope : IIntSampleProvider
    {
        class Segment
        {
            public double TimeFraction { set; get; }
            public double AmplitudeFraction { set; get; }
        }

        class ScaledSegment
        {
            public int SampleCount { set; get; }
            public short Amplitude { set; get; }
        }

        List<Segment> segments = new List<Segment>();
        List<ScaledSegment> scaledSegments = new List<ScaledSegment>();

        double samplesPerMillisecond;
        bool isInEnvelope;
        bool isDamping;
        long sampleCount;
        int dampingAmplitude;
        int dampingAmplitudeDecrease;

        public event EventHandler EnvelopeCompleted;

        public SegmentEnvelope(int sampleRate)
        {
            samplesPerMillisecond = sampleRate / 1000.0;
        }

        public void ClearSegments()
        {
            segments.Clear();
        }

        public void AddSegment(double timeFraction, double amplitudeFraction)
        {
            segments.Add(new Segment()
            {
                TimeFraction = timeFraction,
                AmplitudeFraction = amplitudeFraction
            });
        }

        public void Attack(int timeMultiplier,              // in milliseconds
                           short AmplitudeMultiplier)       
        {
            scaledSegments.Clear();

            // Scale the segments based on the multipliers
            foreach (Segment segment in segments)
            {
                scaledSegments.Add(new ScaledSegment()
                {
                    // Notice sample rate to convert msec to samples!
                    SampleCount = (int)(segment.TimeFraction * samplesPerMillisecond * timeMultiplier),
                    Amplitude = (short)(segment.AmplitudeFraction * AmplitudeMultiplier)
                });
            }

            sampleCount = 0;
            isInEnvelope = true;
            isDamping = false;
        }

        public void Damp(double dampSlope)     // in amplitude values per millisecond
        {
            if (!isInEnvelope)
                return;

            dampingAmplitude = CalculateAmplitude();
            dampingAmplitudeDecrease = (int)(65536 * dampSlope / samplesPerMillisecond);
            isDamping = true;
        }

        public int GetNextIntSample()
        {
            int sample = 0;

            if (isDamping)
            {
                dampingAmplitude -= dampingAmplitudeDecrease;

                if (dampingAmplitude > 0)
                {
                    sample = dampingAmplitude;
                }
                else
                {
                    sample = 0;
                    isInEnvelope = false;
                    isDamping = false;
                    OnEnvelopeCompleted();
                }
            }

            else if (isInEnvelope)
            {
                sample = CalculateAmplitude();
                sampleCount++;
            }

            return sample;
        }

        // Method accesses scaledSegment and sampleCount
        int CalculateAmplitude()
        {
            int sample;
            int fromIndex = -1;
            int toIndex = -1;
            int fromSampleCount = int.MinValue;
            int toSampleCount = int.MaxValue;

            for (int index = 0; index < scaledSegments.Count; index++)
            {
                // Find the "from" segment
                if (sampleCount >= scaledSegments[index].SampleCount &&
                            scaledSegments[index].SampleCount > fromSampleCount)
                {
                    fromSampleCount = scaledSegments[index].SampleCount;
                    fromIndex = index;
                }

                // Find the "to" segment
                if (sampleCount <= scaledSegments[index].SampleCount &&
                            scaledSegments[index].SampleCount < toSampleCount)
                {
                    toSampleCount = scaledSegments[index].SampleCount;
                    toIndex = index;
                }
            }

            if (toIndex == -1 || fromIndex == -1)
            {
                sample = 0;
                isInEnvelope = false;
                OnEnvelopeCompleted();
            }

            else if (toIndex == fromIndex)
            {
                sample = 65536 * scaledSegments[toIndex].Amplitude;
            }
            else
            {
                sample = (int)(65536 * ((sampleCount - fromSampleCount) * scaledSegments[toIndex].Amplitude +
                                        (toSampleCount - sampleCount) * scaledSegments[fromIndex].Amplitude) / (toSampleCount - fromSampleCount));
            }

            return sample;
        }

        protected virtual void OnEnvelopeCompleted()
        {
            if (EnvelopeCompleted != null)
                EnvelopeCompleted(this, EventArgs.Empty);
        }
    }
}
