﻿// TwoFingerManipulation by Charles Petzold, January 2010
using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;

namespace TwoFingerManipulation
{
    public partial class MainPage : UserControl
    {
        // Information for mouse
        bool isTracking;
        Point oldPoint;

        // Information for touch
        List<int> idList = new List<int>();
        Dictionary<int, Point> oldPointDictionary = new Dictionary<int, Point>();

        public MainPage()
        {
            InitializeComponent();
            Touch.FrameReported += OnTouchFrameReported;
        }

        // Mouse logic
        protected override void OnMouseLeftButtonDown(MouseButtonEventArgs args)
        {
            if (args.OriginalSource == txtblk)
            {
                oldPoint = args.GetPosition(LayoutRoot);
                isTracking = true;
                CaptureMouse();
                args.Handled = true;
            }
            base.OnMouseLeftButtonDown(args);
        }

        protected override void OnMouseMove(MouseEventArgs args)
        {
            if (isTracking)
            {
                Point newPoint = args.GetPosition(LayoutRoot);
                Matrix matx = MakeTranslateTransform(newPoint.X - oldPoint.X,
                                                     newPoint.Y - oldPoint.Y);

                xform.Matrix = MultiplyTransforms(xform.Matrix, matx);
                oldPoint = newPoint;
            }
            base.OnMouseMove(args);
        }

        protected override void OnLostMouseCapture(MouseEventArgs e)
        {
            isTracking = false;
            base.OnLostMouseCapture(e);
        }

        // Touch logic
        void OnTouchFrameReported(object sender, TouchFrameEventArgs args)
        {
            TouchPoint primaryTouchPoint = args.GetPrimaryTouchPoint(LayoutRoot);

            // Don't let usable touch points get promoted to mouse events
            if (primaryTouchPoint != null &&
                    primaryTouchPoint.Action == TouchAction.Down &&
                    primaryTouchPoint.TouchDevice.DirectlyOver == txtblk)
                args.SuspendMousePromotionUntilTouchUp();

            // Enumerate all the TouchPoints
            TouchPointCollection touchPoints = args.GetTouchPoints(LayoutRoot);

            foreach (TouchPoint touchPoint in touchPoints)
            {
                TouchDevice touchDevice = touchPoint.TouchDevice;
                int id = touchDevice.Id;

                // Ignore anything beyond two fingers
                if (oldPointDictionary.Count == 2 && !oldPointDictionary.ContainsKey(id))
                    continue;

                switch (touchPoint.Action)
                {
                    case TouchAction.Down:
                        if (oldPointDictionary.ContainsKey(id))
                            throw new Exception("Not expecting a duplicate finger on Down!");

                        if (touchDevice.DirectlyOver != txtblk)
                            break;

                        // Add ID to list and dictionary
                        idList.Add(id);
                        oldPointDictionary.Add(id, touchPoint.Position);
                        break;

                    case TouchAction.Move:
                        if (!oldPointDictionary.ContainsKey(id))
                            break;

                        Point oldPoint = oldPointDictionary[id];
                        Point newPoint = touchPoint.Position;
                        Matrix matx;

                        // For one finger, just do translate
                        if (oldPointDictionary.Count == 1)
                        {
                            matx = MakeTranslateTransform(newPoint.X - oldPoint.X,
                                                          newPoint.Y - oldPoint.Y);
                        }
                        // For two fingers do rotate and scale
                        else
                        {
                            int otherId = idList[1 - idList.IndexOf(id)];
                            Point refPoint = oldPointDictionary[otherId];
                            matx = ComputeMoveMatrix(oldPoint, newPoint, refPoint);
                        }
                        xform.Matrix = MultiplyTransforms(xform.Matrix, matx);
                        oldPointDictionary[id] = newPoint;
                        break;

                    case TouchAction.Up:
                        if (!oldPointDictionary.ContainsKey(id))
                            break;

                        // Remove from list and dictionary
                        idList.Remove(id);
                        oldPointDictionary.Remove(id);
                        break;
                }
            }
        }

        Matrix ComputeMoveMatrix(Point oldPoint, Point newPoint, Point refPoint)
        {
            // Extract rotation component first
            double oldAngle = Math.Atan2(oldPoint.Y - refPoint.Y, oldPoint.X - refPoint.X);
            double newAngle = Math.Atan2(newPoint.Y - refPoint.Y, newPoint.X - refPoint.X);
            Matrix rotate = MakeRotateTransform(newAngle - oldAngle, refPoint.X, refPoint.Y);

            // Now determine scaling beyond any rotation 
            Point rotatedOldPoint = ApplyTransform(rotate, oldPoint);
            double scaleX = (newPoint.X - refPoint.X) / (rotatedOldPoint.X - refPoint.X);
            double scaleY = (newPoint.Y - refPoint.Y) / (rotatedOldPoint.Y - refPoint.Y);
            Matrix scale;
           
            // Don't let bad numbers make the element disappear!
            if (!Double.IsNaN(scaleX) && !Double.IsInfinity(scaleX) && scaleX != 0 &&
                !Double.IsNaN(scaleY) && !Double.IsInfinity(scaleY) && scaleY != 0)
            {
                scale = MakeScaleTransform(scaleX, scaleY, refPoint.X, refPoint.Y);
            }
            else
            {
                scale = Matrix.Identity;
            }
            return MultiplyTransforms(rotate, scale);
        }

        // None of these methods would be required in a WPF app
        Matrix MakeTranslateTransform(double offsetX, double offsetY)
        {
            return new Matrix(1, 0, 0, 1, offsetX, offsetY);
        }

        Matrix MakeScaleTransform(double scaleX, double scaleY)
        {
            return new Matrix(scaleX, 0, 0, scaleY, 0, 0);
        }

        Matrix MakeScaleTransform(double scaleX, double scaleY, double centerX, double centerY)
        {
            Matrix matx = MakeTranslateTransform(-centerX, -centerY);
            matx = MultiplyTransforms(matx, MakeScaleTransform(scaleX, scaleY));
            return MultiplyTransforms(matx, MakeTranslateTransform(centerX, centerY));
        }

        Matrix MakeRotateTransform(double radians)
        {
            double cos = Math.Cos(radians);
            double sin = Math.Sin(radians);
            return new Matrix(cos, sin, -sin, cos, 0, 0);
        }

        Matrix MakeRotateTransform(double radians, double centerX, double centerY)
        {
            Matrix matx = MakeTranslateTransform(-centerX, -centerY);
            matx = MultiplyTransforms(matx, MakeRotateTransform(radians));
            return MultiplyTransforms(matx, MakeTranslateTransform(centerX, centerY));
        }

        Matrix MultiplyTransforms(Matrix A, Matrix B)
        {
            return new Matrix(A.M11 * B.M11 + A.M12 * B.M21,
                              A.M11 * B.M12 + A.M12 * B.M22,
                              A.M21 * B.M11 + A.M22 * B.M21,
                              A.M21 * B.M12 + A.M22 * B.M22,
                              A.OffsetX * B.M11 + A.OffsetY * B.M21 + B.OffsetX,
                              A.OffsetX * B.M12 + A.OffsetY * B.M22 + B.OffsetY);
        }

        Point ApplyTransform(Matrix matx, Point pt)
        {
            return new Point(matx.M11 * pt.X + matx.M21 * pt.Y + matx.OffsetX,
                             matx.M12 * pt.X + matx.M22 * pt.Y + matx.OffsetY);
        }
    }
}
